// femmviewDoc.h : interface of the CFemmviewDoc class
//
/////////////////////////////////////////////////////////////////////////////

class CFemmviewDoc : public CDocument
{
protected: // create from serialization only
	CFemmviewDoc();
	DECLARE_DYNCREATE(CFemmviewDoc)

// Attributes
public:

	// General problem attributes
	double  Frequency;
	int		LengthUnits;
	double *LengthConv;
	BOOL    ProblemType;
	BOOL	Coords;
	CString ProblemNote;
	BOOL	FirstDraw;
	BOOL	Smooth;

	double  A_High, A_Low;
	double	A_lb, A_ub;
	double  Br_High,Br_Low;
	double  Bi_High,Bi_Low;
	double  B_High, B_Low;
	double  B_lb, B_ub;

	// lists of nodes, segments, and block labels
	CArray< CNode, CNode&>             nodelist;
	CArray< CSegment, CSegment&>       linelist;
	CArray< CBlockLabel, CBlockLabel&> blocklist;
	CArray< CArcSegment, CArcSegment&> arclist;

	// CArrays containing the mesh information
	CArray< CMeshNode, CMeshNode&>	   meshnode;
	CArray< CElement, CElement&>	   meshelem;
	
	// List of elements connected to each node;
	int *NumList;
	int **ConList;

	// lists of properties
	CArray< CMaterialProp, CMaterialProp& > blockproplist;
	CArray< CBoundaryProp, CBoundaryProp& > lineproplist;
	CArray< CPointProp,    CPointProp&    > nodeproplist;	
	CArray< CCircuit,      CCircuit&      > circproplist;

	// list of points in a user-defined contour;
	CArray< CComplex, CComplex& > contour;

	// member functions
	int InTriangle(double x, double y);
	BOOL InTriangleTest(double x, double y, int i);
	BOOL GetPointValues(double x, double y, CPointVals &u);
	BOOL GetPointValues(double x, double y, int k, CPointVals &u);
	void GetLineValues(CXYPlot &p, int PlotType, int npoints);
	void GetElementB(CElement &elm);
	void OnReload();
	int ClosestNode(double x, double y);
	CComplex Ctr(int i);
	double ElmArea(int i);
	double ElmArea(CElement *elm);
	void GetPointB(double x, double y, CComplex &B1, CComplex &B2,
		CElement &elm);
	void GetNodalB(CComplex *b1, CComplex *b2,CElement &elm);
	CComplex BlockIntegral(int inttype);
	void LineIntegral(int inttype, CComplex *z);
	int ClosestArcSegment(double x, double y);
	void GetCircle(CArcSegment &asegm,CComplex &c, double &R);
	double ShortestDistanceFromArc(CComplex p, CArcSegment &arc);
	double ShortestDistanceFromSegment(double p, double q, int segm);
	void GetJA(int k,CComplex *J,CComplex *A);
	CComplex PlnInt(double a, CComplex *u, CComplex *v);
	CComplex AxiInt(double a, CComplex *u, CComplex *v,double *r);

/*	// Tests for whether or not one is inside a user-defined
    // contour; this was really mostly used pre- v3.0, when
	// the areas over which block integrals were taken were
	// defined by a closed, user-generated contour.
	BOOL ContourClosed();
	BOOL SlowInContour(double x, double y);
	BOOL InContour(double x, double y);
	BOOL InContour(CElement &elm);
	BOOL InContour(CComplex p);
	BOOL OnContour(CElement &elm);
	void AvgMaxwell(double &fx, double &fy);
	void AvgMaxwell(double &fx,  double &fy,
					CComplex &f2x, CComplex &f2y); 
*/

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFemmviewDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
	virtual BOOL OnOpenDocument(LPCTSTR lpszPathName);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CFemmviewDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CFemmviewDoc)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////
char* StripKey(char *c);