// does Cuthill-McKee algorithm as described in Hoole;

#include<stdafx.h>
#include<afxtempl.h>
#include<stdio.h>
#include<math.h>
#include "fkn.h"
#include "fknDlg.h"
#include "complex.h"
#include "mesh.h"
#include "fullmatrix.h"
#include "spars.h"
#include "FemmeDocCore.h"

BOOL CFemmeDocCore::Cuthill()
{

	FILE *fp;
	int i,j,k,n0,n1,n;
	int newwide,*newnum,**ocon;
	int  *numcon,*nxtnum;
	CNode swap;
	char infile[256];

	// allocate storage
	nxtnum=(int *)calloc(NumNodes,sizeof(int));
	newnum=(int *)calloc(NumNodes,sizeof(int));
	numcon=(int *)calloc(NumNodes,sizeof(int));
	ocon=(int **)calloc(NumNodes,sizeof(int *));
	// initialize node array;
	for(i=0;i<NumNodes;i++)	newnum[i]= -1;

	// read in connectivity from nodefile
	sprintf(infile,"%s.edge",PathName);
	if((fp=fopen(infile,"rt"))==NULL)
	{
		fprintf(stderr,"Couldn't open %s",infile);
		return FALSE;
	}
	fscanf(fp,"%i",&k);	// read in number of lines
	fscanf(fp,"%i",&j);	// read in boundarymarker flag;

	// allocate space for connections;
	ocon[0]=(int *)calloc(2*k,sizeof(int));

	// with first pass, figure out how many connections
	// there are for each node;
	for(i=0;i<k;i++)
	{
		fscanf(fp,"%i",&j);
		fscanf(fp,"%i",&n0);
		fscanf(fp,"%i",&n1);
		fscanf(fp,"%i",&j);

		numcon[n0]++;
		numcon[n1]++;
	}

	// mete out connection storage space;
	for(i=1,n=0;i<NumNodes;i++)
	{
		n+=numcon[i-1];
		ocon[i]=ocon[0]+n;
	}

	// on second pass through file, store connections;
	rewind(fp);
	fscanf(fp,"%li",&k);	// read in number of lines
	fscanf(fp,"%li",&j);	// read in boundarymarker flag;
	for(i=0;i<k;i++)
	{
		fscanf(fp,"%i",&j);
		fscanf(fp,"%i",&n0);
		fscanf(fp,"%i",&n1);
		fscanf(fp,"%i",&j);

		ocon[n0][nxtnum[n0]]=n1; nxtnum[n0]++;
		ocon[n1][nxtnum[n1]]=n0; nxtnum[n1]++;
	}
	fclose(fp);

	// sort connections in order of increasing connectivity;
	// I'm lazy, so I'm doing a bubble sort;
	for(n0=0;n0<NumNodes;n0++)
	{
		for(i=1;i<numcon[n0];i++)
			for(j=1;j<numcon[n0];j++)
				if(numcon[ocon[n0][j]]<numcon[ocon[n0][j-1]])
				{
					n1=ocon[n0][j];
					ocon[n0][j]=ocon[n0][j-1];
					ocon[n0][j-1]=n1;
				}
	}


	// search for a node to start with;
	j=numcon[0]; n0=0;
	for(i=1;i<NumNodes;i++){
		if(numcon[i]<j){
			j=numcon[i];
			n0=i;
		}
		if(j==2) i=k;	// break out if j==2,
						// because this is the best we can do
	}

	// do renumbering algorithm;
	for(i=0;i<NumNodes;i++) nxtnum[i]=-1;
	newnum[n0]=0; n=1; nxtnum[0]=n0;

	do{
		// renumber in order of increasing number of connections;
		
		for(i=0;i<numcon[n0];i++)
		{
			if (newnum[ocon[n0][i]]<0)
			{
				newnum[ocon[n0][i]]=n;
				nxtnum[n]=ocon[n0][i];
				n++;
			}
		}

		// need to catch case in which problem is multiply
		// connected and still renumber right.
		if(nxtnum[newnum[n0]+1]<0){
		//	AfxMessageBox("Multiply Connected!");
		//	exit(0);

			// first, get a node that hasn't been visited yet;
			for(i=0;i<NumNodes;i++)
				if(newnum[i]<0){
					j=numcon[i];
					n0=i;
					i=NumNodes;
				}


			// now, get a new starting node;
			for(i=0;i<NumNodes;i++)
			{
				if((newnum[i]<0) && (numcon[i]<j))
				{
					j=numcon[i];
					n0=i;
				}
				if(j==2) i=NumNodes;	// break out if j==2,
										// because this is the
										// best we can do
			}

			// now, set things to restart;
			newnum[n0]=n;
			nxtnum[n]=n0;
			n++;
		}
		else n0=nxtnum[newnum[n0]+1];

		
	} while(n<NumNodes);

	// remap connectivities;
	for(i=0;i<NumNodes;i++)
		for(j=0;j<numcon[i];j++)
			ocon[i][j]=newnum[ocon[i][j]];

	// remap (anti)periodic boundary points
	for(i=0;i<NumPBCs;i++)
	{
		pbclist[i].x=newnum[pbclist[i].x];
		pbclist[i].y=newnum[pbclist[i].y];
	}

	// find new bandwidth;

	// PBCs fuck up the banding, som could have to do
	// something like:
	// if(NumPBCs!=0) BandWidth=0; 
	// else{
	// but if we apply the PCBs the last thing before the
	// solver is called, we can take advantage of banding
	// speed optimizations without messing things up.
		for(n0=0,newwide=0;n0<NumNodes;n0++)
		{
				for(i=0;i<numcon[n0];i++)
				if(abs(newnum[n0]-ocon[n0][i])>newwide)
				{
					newwide=abs(newnum[n0]-ocon[n0][i]);
				}
		}

		BandWidth=newwide+1;
	// }

	// free up the variables that we needed during the routine....
	free(numcon);
	free(nxtnum);
	free(ocon[0]);
	free(ocon);

	// new mapping remains in newnum;
	// apply this mapping to elements first.
	for(i=0;i<NumEls;i++)
		for(j=0;j<3;j++)
			meshele[i].p[j]=newnum[meshele[i].p[j]];
	
	// now, sort nodes based on newnum;
	for(i=0;i<NumNodes;i++)
		while(newnum[i]!=i)
		{	
			j=newnum[i];
			n=newnum[j]; newnum[j]=newnum[i]; newnum[i]=n;
			swap=meshnode[j]; meshnode[j]=meshnode[i]; meshnode[i]=swap;
		}	
		
	free(newnum);
	return TRUE;
}

CMaterialProp::~CMaterialProp()
{
	if (BHdata!=NULL) free(BHdata);
	if (slope!=NULL)  free(slope);
}

CMaterialProp::CMaterialProp()
{
	slope=NULL;
	BHdata=NULL;
}

void CMaterialProp::GetSlopes()
{
	if (BHpoints==0) return; // catch trivial case;

	CFullMatrix Z;
	int i;
	double l1,l2;

	// take care of LamType=0 situation by changing the
	// apparent B-H curve;
	if(LamType==0)
	{
		for(i=0;i<BHpoints;i++){
			BHdata[i].re = LamFill*BHdata[i].re + 
				           (1.-LamFill)*4.e-07*PI*BHdata[i].im;
		}
	}

	Z.Create(BHpoints);
	Z.Wipe();
	slope=(double *)calloc(BHpoints,sizeof(double));
	
	// smooth out input points to get rid of noise in entered points.
	for(i=1;i<BHpoints-1;i++){
        l1=BHdata[i].re-BHdata[i-1].re;
        l2=BHdata[i+1].re-BHdata[i].re;
        slope[i]=(l2*l2*(BHdata[i-1].im + BHdata[i].im) +
            l1*l2*(BHdata[i-1].im+BHdata[i+1].im) +
            l1*l1*(BHdata[i].im+BHdata[i+1].im))/
            (2*(l1*l1 + l1*l2 + l2*l2));
	}
	for(i=1;i<BHpoints-1;i++) BHdata[i].im=slope[i];

	// impose natural BC on the `left'
	l1=BHdata[1].re-BHdata[0].re;
	Z.M[0][0]=4./l1;
	Z.M[0][1]=2./l1;
	Z.b[0]=6.*(BHdata[1].im-BHdata[0].im)/(l1*l1);

	// impost natural BC on the `right'
	int n=BHpoints;
	l1=BHdata[n-1].re-BHdata[n-2].re;
	Z.M[n-1][n-1]=4./l1;
	Z.M[n-1][n-2]=2./l1;
	Z.b[n-1]=6.*(BHdata[n-1].im-BHdata[n-2].im)/(l1*l1);

	for(i=1;i<BHpoints-1;i++)
	{
		l1=BHdata[i].re-BHdata[i-1].re;
        l2=BHdata[i+1].re-BHdata[i].re;

		Z.M[i][i-1]=2./l1;
		Z.M[i][i]=4.*(l1+l2)/(l1*l2);
		Z.M[i][i+1]=2./l2;

		Z.b[i]=6.*(BHdata[i].im-BHdata[i-1].im)/(l1*l1) +
				 6.*(BHdata[i+1].im-BHdata[i].im)/(l2*l2);
	}

	Z.GaussSolve();

	for(i=0;i<BHpoints;i++) slope[i]=Z.b[i];

	return;
}
