// Solve linear 2-D problems...
#include<stdafx.h>
#include<stdio.h>
#include<math.h>
#include "fkn.h"
#include "fknDlg.h"
#include "complex.h"
#include "mesh.h"
#include "spars.h"
#include "FemmeDocCore.h"
#define Log log

BOOL CFemmeDocCore::HarmonicAxisymmetric(CBigComplexLinProb &L)
{
	int i,j,k,s,flag,sdi_iter,sdin;
	int pctr;
	CComplex Me[3][3],be[3];		// element matrices;
	double l[3],p[3],q[3];		// element shape parameters;
	int n[3];					// numbers of nodes for a particular element;
	double a,r,t,x,y,ds,w,R,rn[3],g[3],a_hat,R_hat;
	CComplex K,mu1,mu2,lag,halflag,deg45,Jv;
	CComplex **Mu,*V_sdi;
	double c=PI*4.e-05;
	double units[]={2.54,0.1,1.,100.,0.00254,1.e-04};
	CElement *El;
	BOOL SDIflag;

	CComplex *CircInt1,*CircInt2,*CircInt3;

	deg45=1+I;
	w=Frequency*2.*PI;


		// check to see if any circuits have been defined and process them;
	if (NumCircProps>0)
	{
		CircInt1=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		CircInt2=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		CircInt3=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		for(i=0;i<NumEls;i++){
			if(meshele[i].lbl>=0)
			if(labellist[meshele[i].lbl].InCircuit!=-1){
				El=&meshele[i];
				
				// get element area;
				for(k=0;k<3;k++) n[k]=El->p[k];
				p[0]=meshnode[n[1]].y - meshnode[n[2]].y;
				p[1]=meshnode[n[2]].y - meshnode[n[0]].y;
				p[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
				q[0]=meshnode[n[2]].x - meshnode[n[1]].x;
				q[1]=meshnode[n[0]].x - meshnode[n[2]].x;
				q[2]=meshnode[n[1]].x - meshnode[n[0]].x;
				a=(p[0]*q[1]-p[1]*q[0])/2.;
				r=(meshnode[n[0]].x+meshnode[n[1]].x+meshnode[n[2]].x)/3.;

			// evaluate integrals;
				
				// total cross-section of circuit;
				CircInt1[labellist[El->lbl].InCircuit]+=a; 

				// integral of conductivity / R over the circuit;
				CircInt2[labellist[El->lbl].InCircuit]+=
					a*blockproplist[El->blk].Cduct/(0.01*r);

				// integral of applied J over current;
				CircInt3[labellist[El->lbl].InCircuit]+=
					(blockproplist[El->blk].Jr+I*blockproplist[El->blk].Ji)*a*100.;
			}
		}

		for(i=0;i<NumCircProps;i++)
		{
			if (circproplist[i].CircType==0) // specified current
			{
				if(CircInt2[i]==0){ //circuit composed of zero cond. materials
					circproplist[i].Case=1;
					if (CircInt1[i]==0.) circproplist[i].J=0.;
					else circproplist[i].J=0.01*(
						(circproplist[i].Amps_re+I*circproplist[i].Amps_im) -
						 CircInt3[i])/CircInt1[i];
				}
				else{
					circproplist[i].Case=2; // need to include an extra
										    // entry in matrix to solve for
					                        // voltage grad in the circuit
				}
			}
			else{
				// case where voltage gradient is specified a priori...
				circproplist[i].Case=0;
				circproplist[i].dV=circproplist[i].dVolts_re +
								   I*circproplist[i].dVolts_im;
			}
		}
	}


	// check to see if there are any SDI boundaries...
	// lineproplist[ meshele[i].e[j] ].BdryFormat==0
	for(i=0;i<NumLineProps;i++)
		if(lineproplist[i].BdryFormat==3) SDIflag=TRUE;

	if(SDIflag==TRUE){
		// there is an SDI boundary defined; check to see if it is in use
		SDIflag=FALSE;
		for(i=0;i<NumEls;i++)
			for(j=0;j<3;j++)
				if (lineproplist[meshele[i].e[j]].BdryFormat==3){
					SDIflag=TRUE;
					printf("Problem has SDI boundaries\n");
					i=NumEls;
					j=3;
				}
	}

	if (SDIflag==TRUE)
	{
		V_sdi=(CComplex *) calloc(NumNodes+NumCircProps,sizeof(CComplex));
		sdin=2;
	}
	else sdin=1;

  for(sdi_iter=0; sdi_iter<sdin; sdi_iter++)
  {
		TheView->SetDlgItemText(IDC_FRAME1,"Matrix Construction");
		TheView->m_prog1.SetPos(0);
		pctr=0;

   if (sdi_iter>0) L.Wipe();
   else{
	// compute effective permeability for each block type;
	Mu=(CComplex **)calloc(NumBlockProps,sizeof(CComplex *));
	for(i=0;i<NumBlockProps;i++) Mu[i]=(CComplex *)calloc(2,sizeof(CComplex));

	for(k=0;k<NumBlockProps;k++){
		lag=exp(-I*blockproplist[k].Theta_h*PI/180.);
		if (blockproplist[k].LamType==0){
			Mu[k][0]=blockproplist[k].mu_x*lag;
			Mu[k][1]=blockproplist[k].mu_y*lag;
			
			if(blockproplist[k].Lam_d!=0){
				halflag=exp(-I*blockproplist[k].Theta_h*PI/360.);

				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_x));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				Mu[k][0]=((Mu[k][0]*tanh(K))/K)*blockproplist[k].LamFill;
	
				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_y));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				Mu[k][1]=((Mu[k][1]*tanh(K))/K)*blockproplist[k].LamFill;
			}
		}
		if (blockproplist[k].LamType==1){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_x;
				Mu[k][0]=(K*x) + (1.-x);
				Mu[k][1]=K/(x + K*(1.-x));
		}
		if (blockproplist[k].LamType==2){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_y;
				Mu[k][1]=(K*x) + (1.-x);
				Mu[k][0]=K/(x + K*(1.-x));

		}
	}
   }

	// build element matrices using the matrices derived in Allaire's book.
	for(i=0;i<NumEls;i++){

		// update ``building matrix'' progress bar...
		j=(i*20)/NumEls+1;
		if(j>pctr){
			j=pctr*5; if (j>100) j=100;
			TheView->m_prog1.SetPos(j);
			pctr++;
		}

		// zero out Me, be;
		for(j=0;j<3;j++){
			for(k=0;k<3;k++) Me[j][k]=0;
			be[j]=0;
			rn[j]=meshnode[meshele[i].p[j]].x;
		}

		// Determine shape parameters.
		// l == element side lengths;
		// p corresponds to the `b' parameter in Allaire
		// q corresponds to the `c' parameter in Allaire
		El=&meshele[i];		
	
		for(k=0;k<3;k++) n[k]=El->p[k];
		p[0]=meshnode[n[1]].y - meshnode[n[2]].y;
		p[1]=meshnode[n[2]].y - meshnode[n[0]].y;
		p[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
		q[0]=meshnode[n[2]].x - meshnode[n[1]].x;
		q[1]=meshnode[n[0]].x - meshnode[n[2]].x;
		q[2]=meshnode[n[1]].x - meshnode[n[0]].x;
		g[0]=(meshnode[n[2]].x + meshnode[n[1]].x)/2.;
		g[1]=(meshnode[n[0]].x + meshnode[n[2]].x)/2.;
		g[2]=(meshnode[n[1]].x + meshnode[n[0]].x)/2.;	

		for(j=0,k=1;j<3;k++,j++){
			if (k==3) k=0;
			l[j]=sqrt( pow(meshnode[n[k]].x-meshnode[n[j]].x,2.) +
					   pow(meshnode[n[k]].y-meshnode[n[j]].y,2.) );
		}
		a=(p[0]*q[1]-p[1]*q[0])/2.;
		R=(meshnode[n[0]].x+meshnode[n[1]].x+meshnode[n[2]].x)/3.;

		for(j=0,a_hat=0;j<3;j++) a_hat+=(rn[j]*rn[j]*p[j]/(4.*R));

		// retrieve permeability for the element;
		k=meshele[i].blk;
		mu1=Mu[k][0];
		mu2=Mu[k][1];	

		for(j=0,flag=0;j<3;j++) if(rn[j]<1.e-06) flag++;
		switch(flag)
		{
			case 2:
				R_hat=R;

			break;

			case 1:

				if(rn[0]<1.e-06){
					if (fabs(rn[1]-rn[2])<1.e-06) R_hat=rn[2]/2.;
					else R_hat=(rn[1] - rn[2])/(2.*log(rn[1]) - 2.*log(rn[2]));
				}
				if(rn[1]<1.e-06){
					if (fabs(rn[2]-rn[0])<1.e-06) R_hat=rn[0]/2.;
					else R_hat=(rn[2] - rn[0])/(2.*log(rn[2]) - 2.*log(rn[0]));
				}
				if(rn[2]<1.e-06){
					if (fabs(rn[0]-rn[1])<1.e-06) R_hat=rn[1]/2.;
					else R_hat=(rn[0] - rn[1])/(2.*log(rn[0]) - 2.*log(rn[1]));
				}

			break;

			default:

				if (fabs(q[0])<1.e-06)
					R_hat=(q[1]*q[1])/(2.*(-q[1] + rn[0]*log(rn[0]/rn[2])));
				else if (fabs(q[1])<1.e-06)
					R_hat=(q[2]*q[2])/(2.*(-q[2] + rn[1]*log(rn[1]/rn[0])));
				else if (fabs(q[2])<1.e-06)
					R_hat=(q[0]*q[0])/(2.*(-q[0] + rn[2]*log(rn[2]/rn[1])));
			    else
					R_hat=-(q[0]*q[1]*q[2])/
						   (2.*(q[0]*rn[0]*log(rn[0]) + 
						        q[1]*rn[1]*log(rn[1]) + 
								q[2]*rn[2]*log(rn[2])));

			break;
		}

		// Mr Contribution
		// Derived from flux formulation with c0 + c1 r^2 + c2 z 
		// interpolation in the element.
		K=(-1./(2.*a_hat*R*mu2));
		for(j=0;j<3;j++)
			for(k=j;k<3;k++)
				Me[j][k] += K*p[j]*rn[j]*p[k]*rn[k];

		// need this loop to avoid singularities.  This just puts something
		// on the main diagonal of nodes that are on the r=0 line.
		// The program later sets these nodes to zero, but it's good to
		// for scaling reasons to grab entries from the neighboring diagonals
		// rather than just setting these entries to 1 or something....
		for(j=0;j<3;j++)
			if (rn[j]<1.e-06) Me[j][j]+=Me[0][0]+Me[1][1]+Me[2][2];

		// Mz Contribution;  
		// Derived from flux formulation with c0 + c1 r^2 + c2 z 
		// interpolation in the element.
		K=(-1./(2.*a_hat*R_hat*mu1));
		for(j=0;j<3;j++)
			for(k=j;k<3;k++)
				Me[j][k] += K*(q[j]*rn[j])*(q[k]*rn[k])*
			                  (g[j]/R)*(g[k]/R);

		// contribution from eddy currents;
		// This is from ``old'' formulation;
		// Does it need changes to work `right' with the
		// new formulation?
		K.Set(0.,-2.*R*a*w*blockproplist[meshele[i].blk].Cduct*c/12.);
		if((blockproplist[El->blk].LamType==0) &&
			(blockproplist[El->blk].Lam_d>0)) K=0;
		for(j=0;j<3;j++)
			for(k=j;k<3;k++){
				Me[j][k]+=K;
				if(j==k) Me[j][k]+=K;
			}
			

		// contributions to Me, be from derivative boundary conditions;
		for(j=0;j<3;j++){
			k=j+1; if(k==3) k=0;
			r=(meshnode[n[j]].x+meshnode[n[k]].x)/2.;
			if (El->e[j] >= 0)
			if (lineproplist[El->e[j]].BdryFormat==2)
			{
				// conversion factor is 10^(-4) (I think...)
				
				K.Set(-0.0001*c*2.*r*lineproplist[ El->e[j] ].c0*l[j]/6.,0);
				Me[j][j]+=2*K;
				Me[k][k]+=2*K;
				Me[j][k]+=K;
				Me[k][j]+=K;

				K.Set((lineproplist[ El->e[j] ].c1*l[j]/2.)*2.*r*0.0001,0);
				be[j]+=K;
				be[k]+=K;
			}
			
			if (lineproplist[El->e[j]].BdryFormat==1)
			{
				ds=sqrt(2./(0.4*PI*w*lineproplist[El->e[j]].Sig*
					lineproplist[El->e[j]].Mu));
				K=deg45/(-ds*lineproplist[El->e[j]].Mu*100.);
				K*=(2.*r*l[j]/6.);
				Me[j][j]+=2*K;
				Me[k][k]+=2*K;
				Me[j][k]+=K;
				Me[k][j]+=K;
			}
		}
/*		
		// contribution to be from current density in the block
		for(j=0;j<3;j++){
			K.Set(blockproplist[El->blk].Jr*(3.*R+rn[j])*a/6.,
				  blockproplist[El->blk].Ji*(3.*R+rn[j])*a/6.);
			be[j]-=K;
		}
*/

		// contribution to be from current density in the block
		for(j=0;j<3;j++){
			Jv=0;
			if(labellist[El->lbl].InCircuit>=0)
			{
				k=labellist[El->lbl].InCircuit;
				if(circproplist[k].Case==1) Jv=circproplist[k].J;
				if(circproplist[k].Case==0)
					Jv=-100.*circproplist[k].dV*
					    blockproplist[El->blk].Cduct/R;
			}

			K=-2.*R*(blockproplist[El->blk].Jr+
			   I*blockproplist[El->blk].Ji+Jv)*a/3.;
			be[j]+=K;

			if(labellist[El->lbl].InCircuit>=0){
				k=labellist[El->lbl].InCircuit;
				if(circproplist[k].Case==2)
					L.b[NumNodes+k]+=K/R;
			}
		}

		// do Case 2 circuit stuff for element
		if(labellist[El->lbl].InCircuit>=0){
			k=labellist[El->lbl].InCircuit;
			if(circproplist[k].Case==2){
				K=-2.*I*a*w*blockproplist[meshele[i].blk].Cduct*c;
				for(j=0;j<3;j++) L.Put(L.Get(n[j],NumNodes+k)+K/3.,n[j],NumNodes+k);
				L.Put(L.Get(NumNodes+k,NumNodes+k)+K/R,NumNodes+k,NumNodes+k);
			}
		}

		// combine block matrices into global matrices;
		for (j=0;j<3;j++){
			for (k=j;k<3;k++)
				L.Put( L.Get(n[j],n[k])+Me[j][k], n[j],n[k]);
			L.b[n[j]]+=be[j];
		}
	}

	// add in contribution from point currents;
	for(i=0;i<NumNodes;i++)
		if(meshnode[i].bc>=0){
			r=meshnode[i].x;
			K.Set(2.*r*0.01*nodeproplist[meshnode[i].bc].Jr,
				  2.*r*0.01*nodeproplist[meshnode[i].bc].Ji);
			L.b[i]-=K;
		}

	// add in total current constraints for circuits;
	for(i=0;i<NumCircProps;i++)
		if (circproplist[i].Case==2){
			L.b[NumNodes+i]+=2.*0.01*(circproplist[i].Amps_re +
				                   I*circproplist[i].Amps_im);
		}
	
	// apply fixed boundary conditions at points;
	for(i=0;i<NumNodes;i++)
		if(meshnode[i].x<(units[LengthUnits]*1.e-06)){
			K=0;
			L.SetValue(i,K);
		}
		else if(meshnode[i].bc >=0)
			if((nodeproplist[meshnode[i].bc].Jr==0) &&
			   (nodeproplist[meshnode[i].bc].Ji==0) && (sdi_iter==0))
			{
				K.Set(nodeproplist[meshnode[i].bc].Ar / c,
					  nodeproplist[meshnode[i].bc].Ai / c);
				L.SetValue(i,K);
			}

	// apply fixed boundary conditions along segments;
	for(i=0;i<NumEls;i++)
		for(j=0;j<3;j++){
			k=j+1; if(k==3) k=0;
			if(meshele[i].e[j]>=0)
			if(lineproplist[ meshele[i].e[j] ].BdryFormat==0)
			{
				if(Coords==0){
					// first point on the side;
					x=meshnode[meshele[i].p[j]].x;
					y=meshnode[meshele[i].p[j]].y;
					x*=units[LengthUnits]; y*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + x*lineproplist[s].A1 +
					  y*lineproplist[s].A2;
					// convert into appropriate imaginary number.
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[j],K);
					
					// second point on the side;
					x=meshnode[meshele[i].p[k]].x;
					y=meshnode[meshele[i].p[k]].y;
					x*=units[LengthUnits]; y*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + x*lineproplist[s].A1 +
					  y*lineproplist[s].A2;
					// just take``real'' component.
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[k],K);
				}
				else{
					// first point on the side;
					x=meshnode[meshele[i].p[j]].x;
					y=meshnode[meshele[i].p[j]].y;
					r=sqrt(x*x+y*y);
					if ((x==0) && (y==0)) t=0;
					else t=atan2(y,x)*180./PI;
					r*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + r*lineproplist[s].A1 +
					  t*lineproplist[s].A2;
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[j],K);
					
					// second point on the side;
					x=meshnode[meshele[i].p[k]].x;
					y=meshnode[meshele[i].p[k]].y;
					r=sqrt(x*x+y*y);
					if((x==0) && (y==0)) t=0;
					else t=atan2(y,x)*180./PI;
					r*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + r*lineproplist[s].A1 +
					  t*lineproplist[s].A2;
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[k],K);
				}

			}
		}
	
	if ((SDIflag==TRUE) && (sdi_iter==1)) for(i=0;i<NumEls;i++)
		for(j=0;j<3;j++){
			k=j+1; if(k==3) k=0;
			if(meshele[i].e[j]>=0)
			if(lineproplist[ meshele[i].e[j] ].BdryFormat==3)
			{			
					L.SetValue(meshele[i].p[j],0.*I);
					L.SetValue(meshele[i].p[k],0.*I);
			}
		}

	// "fix" diagonal entries associated with circuits that have 
	// applied current or voltage that is known a priori
	// so that solver doesn't throw a "singular" flag
	for(j=0;j<NumCircProps;j++)
		if (circproplist[j].Case<2)	L.Put(L.Get(0,0),NumNodes+j,NumNodes+j);
	
	// Apply any periodicity/antiperiodicity boundary conditions that we have
	if (NumPBCs>0)
	{
		TheView->SetDlgItemText(IDC_FRAME1,"Applying (Anti)Periodic BCs");
		TheView->m_prog1.SetPos(0);
		pctr=0;	
		
	}
	for(k=0;k<NumPBCs;k++)
	{
		if (pbclist[k].t==0) L.Periodicity(pbclist[k].x,pbclist[k].y);
		if (pbclist[k].t==1) L.AntiPeriodicity(pbclist[k].x,pbclist[k].y);

		// update progress bar...
		j=(k*20)/NumPBCs+1;
		if(j>pctr){
			j=pctr*5; if (j>100) j=100;
			TheView->m_prog1.SetPos(j);
			pctr++;
		}
	}

	// solve the problem;
	fprintf(stderr,"Solving...\n");
	if (L.PBCGSolve(FALSE)==FALSE) return FALSE;

	if ((SDIflag==TRUE) && (sdi_iter==0))
		for (i=0;i<NumNodes+NumCircProps;i++) V_sdi[i]=L.V[i];
	if (sdi_iter==1)
		for (i=0;i<NumNodes+NumCircProps;i++)
			L.V[i]=(V_sdi[i]+L.V[i])/2.;

} //end of SDI loop;
	// convert answer back to webers
	for (i=0;i<NumNodes;i++) L.b[i]=L.V[i]*c*2.*PI*meshnode[i].x*0.01;
	for (i=0;i<NumCircProps;i++)
		L.b[NumNodes+i]=(I*w*c*0.01*L.V[NumNodes+i]);
	

	// free up space allocated in this routine
	for(k=0;k<NumBlockProps;k++) free(Mu[k]);
	free(Mu);
	if (SDIflag==TRUE) free(V_sdi);
	if(NumCircProps>0){
		free(CircInt1);
		free(CircInt2);
		free(CircInt3);
	}
	return TRUE;
}
